@@ -5,9 +5,10 @@
 
 #include "sd-daemon.h"
 
+#include "build.h"
 #include "conf-parser.h"
+#include "constants.h"
 #include "daemon-util.h"
-#include "def.h"
 #include "fd-util.h"
 #include "fileio.h"
 #include "journal-remote-write.h"
@@ -15,9 +16,11 @@
 #include "main-func.h"
 #include "memory-util.h"
 #include "parse-argument.h"
+#include "parse-helpers.h"
 #include "pretty-print.h"
 #include "process-util.h"
 #include "rlimit-util.h"
+#include "sigbus.h"
 #include "signal-util.h"
 #include "socket-netlink.h"
 #include "socket-util.h"
@@ -41,7 +44,7 @@ static int http_socket = -1, https_socket = -1;
 static char** arg_gnutls_log = NULL;
 
 static JournalWriteSplitMode arg_split_mode = _JOURNAL_WRITE_SPLIT_INVALID;
-static const char* arg_output = NULL;
+static char *arg_output = NULL;
 
 static char *arg_key = NULL;
 static char *arg_cert = NULL;
@@ -52,10 +55,16 @@ static bool arg_trust_all = false;
 static bool arg_trust_all = true;
 #endif
 
+static uint64_t arg_max_use = UINT64_MAX;
+static uint64_t arg_max_size = UINT64_MAX;
+static uint64_t arg_n_max_files = UINT64_MAX;
+static uint64_t arg_keep_free = UINT64_MAX;
+
 STATIC_DESTRUCTOR_REGISTER(arg_gnutls_log, strv_freep);
 STATIC_DESTRUCTOR_REGISTER(arg_key, freep);
 STATIC_DESTRUCTOR_REGISTER(arg_cert, freep);
 STATIC_DESTRUCTOR_REGISTER(arg_trust, freep);
+STATIC_DESTRUCTOR_REGISTER(arg_output, freep);
 
 static const char* const journal_write_split_mode_table[_JOURNAL_WRITE_SPLIT_MAX] = {
         [JOURNAL_WRITE_SPLIT_NONE] = "none",
@@ -79,7 +88,10 @@ static int spawn_child(const char* child, char** argv) {
         if (pipe(fd) < 0)
                 return log_error_errno(errno, "Failed to create pager pipe: %m");
 
-        r = safe_fork("(remote)", FORK_RESET_SIGNALS|FORK_DEATHSIG|FORK_LOG, &child_pid);
+        r = safe_fork_full("(remote)",
+                           (int[]) {STDIN_FILENO, fd[1], STDERR_FILENO },
+                           NULL, 0,
+                           FORK_RESET_SIGNALS|FORK_CLOSE_ALL_FDS|FORK_DEATHSIG|FORK_REARRANGE_STDIO|FORK_LOG|FORK_RLIMIT_NOFILE_SAFE, &child_pid);
         if (r < 0) {
                 safe_close_pair(fd);
                 return r;
@@ -87,16 +99,6 @@ static int spawn_child(const char* child, char** argv) {
 
         /* In the child */
         if (r == 0) {
-                fd[0] = safe_close(fd[0]);
-
-                r = rearrange_stdio(STDIN_FILENO, TAKE_FD(fd[1]), STDERR_FILENO);
-                if (r < 0) {
-                        log_error_errno(r, "Failed to dup pipe to stdout: %m");
-                        _exit(EXIT_FAILURE);
-                }
-
-                (void) rlimit_nofile_safe();
-
                 execvp(child, argv);
                 log_error_errno(errno, "Failed to exec child %s: %m", child);
                 _exit(EXIT_FAILURE);
@@ -223,9 +225,7 @@ static int process_http_upload(
                 finished = true;
 
         for (;;) {
-                r = process_source(source,
-                                   journal_remote_server_global->compress,
-                                   journal_remote_server_global->seal);
+                r = process_source(source, journal_remote_server_global->file_flags);
                 if (r == -EAGAIN)
                         break;
                 if (r < 0) {
@@ -312,7 +312,7 @@ static mhd_result request_handler(
 
                 if (chunked)
                         return mhd_respond(connection, MHD_HTTP_BAD_REQUEST,
-                                           "Content-Length must not specified when Transfer-Encoding type is 'chuncked'");
+                                           "Content-Length not allowed when Transfer-Encoding type is 'chunked'");
 
                 r = safe_atozu(header, &len);
                 if (r < 0)
@@ -371,8 +371,8 @@ static int setup_microhttpd_server(RemoteServer *s,
                                    const char *cert,
                                    const char *trust) {
         struct MHD_OptionItem opts[] = {
-                { MHD_OPTION_NOTIFY_COMPLETED, (intptr_t) request_meta_free},
                 { MHD_OPTION_EXTERNAL_LOGGER, (intptr_t) microhttpd_logger},
+                { MHD_OPTION_NOTIFY_COMPLETED, (intptr_t) request_meta_free},
                 { MHD_OPTION_LISTEN_SOCKET, fd},
                 { MHD_OPTION_CONNECTION_MEMORY_LIMIT, 128*1024},
                 { MHD_OPTION_END},
@@ -387,9 +387,9 @@ static int setup_microhttpd_server(RemoteServer *s,
                 MHD_USE_EPOLL |
                 MHD_USE_ITC;
 
+        _cleanup_(MHDDaemonWrapper_freep) MHDDaemonWrapper *d = NULL;
         const union MHD_DaemonInfo *info;
         int r, epoll_fd;
-        MHDDaemonWrapper *d;
 
         assert(fd >= 0);
 
@@ -437,75 +437,49 @@ static int setup_microhttpd_server(RemoteServer *s,
                                      request_handler, NULL,
                                      MHD_OPTION_ARRAY, opts,
                                      MHD_OPTION_END);
-        if (!d->daemon) {
-                log_error("Failed to start µhttp daemon");
-                r = -EINVAL;
-                goto error;
-        }
+        if (!d->daemon)
+                return log_error_errno(SYNTHETIC_ERRNO(EINVAL), "Failed to start μhttp daemon");
 
         log_debug("Started MHD %s daemon on fd:%d (wrapper @ %p)",
                   key ? "HTTPS" : "HTTP", fd, d);
 
         info = MHD_get_daemon_info(d->daemon, MHD_DAEMON_INFO_EPOLL_FD_LINUX_ONLY);
-        if (!info) {
-                log_error("µhttp returned NULL daemon info");
-                r = -EOPNOTSUPP;
-                goto error;
-        }
+        if (!info)
+                return log_error_errno(SYNTHETIC_ERRNO(EOPNOTSUPP), "μhttp returned NULL daemon info");
 
         epoll_fd = info->listen_fd;
-        if (epoll_fd < 0) {
-                log_error("µhttp epoll fd is invalid");
-                r = -EUCLEAN;
-                goto error;
-        }
+        if (epoll_fd < 0)
+                return log_error_errno(SYNTHETIC_ERRNO(EUCLEAN), "μhttp epoll fd is invalid");
 
         r = sd_event_add_io(s->events, &d->io_event,
                             epoll_fd, EPOLLIN,
                             dispatch_http_event, d);
-        if (r < 0) {
-                log_error_errno(r, "Failed to add event callback: %m");
-                goto error;
-        }
+        if (r < 0)
+                return log_error_errno(r, "Failed to add event callback: %m");
 
         r = sd_event_source_set_description(d->io_event, "io_event");
-        if (r < 0) {
-                log_error_errno(r, "Failed to set source name: %m");
-                goto error;
-        }
+        if (r < 0)
+                return log_error_errno(r, "Failed to set source name: %m");
 
         r = sd_event_add_time(s->events, &d->timer_event,
                               CLOCK_MONOTONIC, UINT64_MAX, 0,
                               null_timer_event_handler, d);
-        if (r < 0) {
-                log_error_errno(r, "Failed to add timer_event: %m");
-                goto error;
-        }
+        if (r < 0)
+                return log_error_errno(r, "Failed to add timer_event: %m");
 
         r = sd_event_source_set_description(d->timer_event, "timer_event");
-        if (r < 0) {
-                log_error_errno(r, "Failed to set source name: %m");
-                goto error;
-        }
+        if (r < 0)
+                return log_error_errno(r, "Failed to set source name: %m");
 
         r = hashmap_ensure_put(&s->daemons, &uint64_hash_ops, &d->fd, d);
-        if (r == -ENOMEM) {
-                log_oom();
-                goto error;
-        }
-        if (r < 0) {
-                log_error_errno(r, "Failed to add daemon to hashmap: %m");
-                goto error;
-        }
+        if (r == -ENOMEM)
+                return log_oom();
+        if (r < 0)
+                return log_error_errno(r, "Failed to add daemon to hashmap: %m");
 
+        TAKE_PTR(d);
         s->active++;
         return 0;
-
-error:
-        MHD_stop_daemon(d->daemon);
-        free(d->daemon);
-        free(d);
-        return r;
 }
 
 static int setup_microhttpd_socket(RemoteServer *s,
@@ -532,12 +506,10 @@ static int dispatch_http_event(sd_event_source *event,
                                int fd,
                                uint32_t revents,
                                void *userdata) {
-        MHDDaemonWrapper *d = userdata;
+        MHDDaemonWrapper *d = ASSERT_PTR(userdata);
         int r;
         MHD_UNSIGNED_LONG_LONG timeout = ULLONG_MAX;
 
-        assert(d);
-
         r = MHD_run(d->daemon);
         if (r == MHD_NO)
                 // FIXME: unregister daemon
@@ -598,9 +570,13 @@ static int create_remoteserver(
                 const char* trust) {
 
         int r, n, fd;
-        char **file;
 
-        r = journal_remote_server_init(s, arg_output, arg_split_mode, arg_compress, arg_seal);
+        r = journal_remote_server_init(
+                        s,
+                        arg_output,
+                        arg_split_mode,
+                        (arg_compress ? JOURNAL_COMPRESS : 0) |
+                        (arg_seal ? JOURNAL_SEAL : 0));
         if (r < 0)
                 return r;
 
@@ -735,7 +711,7 @@ static int create_remoteserver(
                    create output as expected. */
                 r = journal_remote_get_writer(s, NULL, &s->_single_writer);
                 if (r < 0)
-                        return r;
+                        return log_warning_errno(r, "Failed to get writer: %m");
         }
 
         return 0;
@@ -758,22 +734,21 @@ static int negative_fd(const char *spec) {
 
 static int parse_config(void) {
         const ConfigTableItem items[] = {
-                { "Remote",  "Seal",                   config_parse_bool,             0, &arg_seal       },
-                { "Remote",  "SplitMode",              config_parse_write_split_mode, 0, &arg_split_mode },
-                { "Remote",  "ServerKeyFile",          config_parse_path,             0, &arg_key        },
-                { "Remote",  "ServerCertificateFile",  config_parse_path,             0, &arg_cert       },
-                { "Remote",  "TrustedCertificateFile", config_parse_path,             0, &arg_trust      },
+                { "Remote",  "Seal",                   config_parse_bool,             0, &arg_seal        },
+                { "Remote",  "SplitMode",              config_parse_write_split_mode, 0, &arg_split_mode  },
+                { "Remote",  "ServerKeyFile",          config_parse_path,             0, &arg_key         },
+                { "Remote",  "ServerCertificateFile",  config_parse_path,             0, &arg_cert        },
+                { "Remote",  "TrustedCertificateFile", config_parse_path_or_ignore,   0, &arg_trust       },
+                { "Remote",  "MaxUse",                 config_parse_iec_uint64,       0, &arg_max_use     },
+                { "Remote",  "MaxFileSize",            config_parse_iec_uint64,       0, &arg_max_size    },
+                { "Remote",  "MaxFiles",               config_parse_uint64,           0, &arg_n_max_files },
+                { "Remote",  "KeepFree",               config_parse_iec_uint64,       0, &arg_keep_free   },
                 {}
         };
 
-        return config_parse_many_nulstr(
-                        PKGSYSCONFDIR "/journal-remote.conf",
-                        CONF_PATHS_NULSTR("systemd/journal-remote.conf.d"),
-                        "Remote\0",
-                        config_item_table_lookup, items,
-                        CONFIG_PARSE_WARN,
-                        NULL,
-                        NULL);
+        return config_parse_config_file("journal-remote.conf", "Remote\0",
+                                        config_item_table_lookup, items,
+                                        CONFIG_PARSE_WARN, NULL);
 }
 
 static int help(void) {
@@ -856,7 +831,7 @@ static int parse_argv(int argc, char *argv[]) {
         assert(argv);
 
         while ((c = getopt_long(argc, argv, "ho:", options, NULL)) >= 0)
-                switch(c) {
+                switch (c) {
 
                 case 'h':
                         return help();
@@ -867,7 +842,7 @@ static int parse_argv(int argc, char *argv[]) {
                 case ARG_URL:
                         if (arg_url)
                                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                                       "cannot currently set more than one --url");
+                                                       "Cannot currently set more than one --url=");
 
                         arg_url = optarg;
                         break;
@@ -875,7 +850,7 @@ static int parse_argv(int argc, char *argv[]) {
                 case ARG_GETTER:
                         if (arg_getter)
                                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                                       "cannot currently use --getter more than once");
+                                                       "Cannot currently use --getter= more than once");
 
                         arg_getter = optarg;
                         break;
@@ -883,7 +858,7 @@ static int parse_argv(int argc, char *argv[]) {
                 case ARG_LISTEN_RAW:
                         if (arg_listen_raw)
                                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                                       "cannot currently use --listen-raw more than once");
+                                                       "Cannot currently use --listen-raw= more than once");
 
                         arg_listen_raw = optarg;
                         break;
@@ -891,7 +866,7 @@ static int parse_argv(int argc, char *argv[]) {
                 case ARG_LISTEN_HTTP:
                         if (arg_listen_http || http_socket >= 0)
                                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                                       "cannot currently use --listen-http more than once");
+                                                       "Cannot currently use --listen-http= more than once");
 
                         r = negative_fd(optarg);
                         if (r >= 0)
@@ -903,7 +878,7 @@ static int parse_argv(int argc, char *argv[]) {
                 case ARG_LISTEN_HTTPS:
                         if (arg_listen_https || https_socket >= 0)
                                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                                       "cannot currently use --listen-https more than once");
+                                                       "Cannot currently use --listen-https= more than once");
 
                         r = negative_fd(optarg);
                         if (r >= 0)
@@ -937,29 +912,27 @@ static int parse_argv(int argc, char *argv[]) {
 
                 case ARG_TRUST:
 #if HAVE_GNUTLS
-                        if (arg_trust || arg_trust_all)
+                        if (arg_trust)
                                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                                       "Confusing trusted CA configuration");
+                                                       "Cannot use --trust more= than once");
 
-                        if (streq(optarg, "all"))
-                                arg_trust_all = true;
-                        else {
-                                arg_trust = strdup(optarg);
-                                if (!arg_trust)
-                                        return log_oom();
-                        }
+                        arg_trust = strdup(optarg);
+                        if (!arg_trust)
+                                return log_oom();
 #else
                         return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                               "Option --trust is not available.");
+                                               "Option --trust= is not available.");
 #endif
                         break;
 
                 case 'o':
                         if (arg_output)
                                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                                       "cannot use --output/-o more than once");
+                                                       "Cannot use --output=/-o more than once");
 
-                        arg_output = optarg;
+                        r = parse_path_argument(optarg, /* suppress_root = */ false, &arg_output);
+                        if (r < 0)
+                                return r;
                         break;
 
                 case ARG_SPLIT_MODE:
@@ -999,7 +972,7 @@ static int parse_argv(int argc, char *argv[]) {
                         break;
 #else
                         return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                               "Option --gnutls-log is not available.");
+                                               "Option --gnutls-log= is not available.");
 #endif
 
                 case '?':
@@ -1019,12 +992,12 @@ static int parse_argv(int argc, char *argv[]) {
                 || sd_listen_fds(false) > 0;
         if (type_a && type_b)
                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                       "Cannot use file input or --getter with "
-                                       "--arg-listen-... or socket activation.");
+                                       "Cannot use file input or --getter= with "
+                                       "--listen-...= or socket activation.");
         if (type_a) {
                 if (!arg_output)
                         return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                               "Option --output must be specified with file input or --getter.");
+                                               "Option --output= must be specified with file input or --getter=.");
 
                 if (!IN_SET(arg_split_mode, JOURNAL_WRITE_SPLIT_NONE, _JOURNAL_WRITE_SPLIT_INVALID))
                         return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
@@ -1050,6 +1023,11 @@ static int parse_argv(int argc, char *argv[]) {
                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
                                        "For SplitMode=host, output must be a directory.");
 
+        if (STRPTR_IN_SET(arg_trust, "-", "all")) {
+                arg_trust_all = true;
+                arg_trust = mfree(arg_trust);
+        }
+
         log_debug("Full config: SplitMode=%s Key=%s Cert=%s Trust=%s",
                   journal_write_split_mode_to_string(arg_split_mode),
                   strna(arg_key),
@@ -1095,7 +1073,7 @@ static int load_certificates(char **key, char **cert, char **trust) {
 
         if ((arg_listen_raw || arg_listen_http) && *trust)
                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
-                                       "Option --trust makes all non-HTTPS connections untrusted.");
+                                       "Option --trust= makes all non-HTTPS connections untrusted.");
 
         return 0;
 }
@@ -1113,6 +1091,8 @@ static int run(int argc, char **argv) {
         /* The journal merging logic potentially needs a lot of fds. */
         (void) rlimit_nofile_bump(HIGH_RLIMIT_NOFILE);
 
+        sigbus_install();
+
         r = parse_config();
         if (r < 0)
                 return r;
@@ -1135,6 +1115,12 @@ static int run(int argc, char **argv) {
                 s.check_trust = !arg_trust_all;
         }
 
+        journal_reset_metrics(&s.metrics);
+        s.metrics.max_use = arg_max_use;
+        s.metrics.max_size = arg_max_size;
+        s.metrics.keep_free = arg_keep_free;
+        s.metrics.n_max_files = arg_n_max_files;
+
         r = create_remoteserver(&s, key, cert, trust);
         if (r < 0)
                 return r;
